import stripe
from contextlib import contextmanager


class InstrumentedClient(stripe.http_client.HTTPClient):
    def __init__(self, client):
        self.client = client
        self.calls = []

    def __getattr__(self, name):
        return getattr(self.client, name)

    def request(self, method, url, headers, post_data=None):
        data = [method, url, -1, None, None]
        try:
            content, status_code, headers = self.client.request(method, url, headers, post_data)
            data[2] = status_code
            if 'Request-Id' in headers:
                data[3] = headers['Request-Id']
        except Exception as err:
            data[4] = str(err)
            raise err
        finally:
            self.calls.append(data)
        return content, status_code, headers


@contextmanager
def instrument_client(client=None):
    from stripe import verify_ssl_certs as verify
    from stripe import proxy

    previous_client = client or stripe.default_http_client or \
        stripe.http_client.new_default_http_client(
            verify_ssl_certs=verify, proxy=proxy)

    try:
        instrumented_client = InstrumentedClient(previous_client)
        stripe.default_http_client = instrumented_client
        yield instrumented_client.calls
    finally:
        stripe.default_http_client = previous_client



import factory
from . import models


class CustomerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Customer
        exclude = 'default_source', 'sources',

    id = factory.Faker('md5', raw_output=False)
    delinquent = False
    default_source = None
    sources = []

    data = factory.Dict({
        "object": "customer",
        "id": factory.SelfAttribute('..id'),
        "account_balance": 0,
        "currency": "usd",
        "delinquent": factory.SelfAttribute('..delinquent'),
        "default_source": factory.SelfAttribute('..default_source'),
        "sources": factory.Dict({
            "data": factory.SelfAttribute('...sources'),
        })
    })


class SubscriptionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Subscription

    id = factory.Faker('md5', raw_output=False)
    customer = factory.SubFactory('rayures.factories.CustomerFactory')
    plan = factory.SubFactory('rayures.factories.PlanFactory')
    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "object": "subscription",
        "customer": factory.SelfAttribute('..customer.id'),
        "plan": factory.Dict({
            "id": factory.SelfAttribute('...plan.id'),
            "object": "plan",
        }),
        "items": factory.Dict({
            "object": "list",
            "data": factory.List([
                factory.Dict({
                    "id": factory.Faker('md5', raw_output=False),
                    "object": "subscription_item",
                    "plan": factory.Dict({
                        "id": factory.SelfAttribute('....plan.id'),
                        "object": "plan",
                    })
                })
            ])
        })
    })


class PlanFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Plan

    id = factory.Faker('md5', raw_output=False)
    product = factory.SubFactory("rayures.factories.ProductFactory", type="service")
    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "object": "plan",
        "product": factory.SelfAttribute('..product.id'),
    })


class ProductFactory(factory.Factory):
    class Meta:
        model = models.Product

    id = factory.Faker('md5', raw_output=False)
    type = "service"
    data = factory.Dict({
        "object": "product",
        "id": factory.SelfAttribute('..id'),
        "type": factory.SelfAttribute('..type'),
    })


class CardFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = models.Card

    id = factory.Faker('md5', raw_output=False)
    fingerprint = factory.Faker('md5', raw_output=False)
    customer = factory.SubFactory("rayures.factories.CustomerFactory")
    data = factory.Dict({
        "id": factory.SelfAttribute('..id'),
        "name": None,
        "brand": "Visa",
        "last4": "3956",
        "object": "card",
        "country": "US",
        "funding": "credit",
        "customer": factory.SelfAttribute('..customer.id'),
        "exp_year": 2018,
        "metadata": {},
        "cvc_check": "pass",
        "exp_month": 10,
        "fingerprint": factory.SelfAttribute('..fingerprint'),
        "tokenization_method": None
    })



from .models import PersistedModel
# from django.db.models.signals import post_init
from django.db.models.signals import post_save, pre_delete


def update_persisted(sender, instance, **kwargs):
    instance.persisted = True
    instance.deleted = False


def delete_persisted(sender, instance, **kwargs):
    instance.persisted = False
    instance.deleted = True


# def init_persisted(sender, instance, **kwargs):
#     instance.persisted = None
#     instance.deleted = instance.deleted_at is not None


for model in PersistedModel.__subclasses__():
    post_save.connect(update_persisted, sender=model)
    pre_delete.connect(delete_persisted, sender=model)
    # post_init.connect(init_persisted, sender=model)



class BaseCustomerFinder:
    def find(self, request):
        """Find customer based on request.

        Must return a rayure.Customer instance or None
        """
        raise NotImplementedError




class DispatchException(Exception):
    def __init__(self, message, *, process, event, proc_errors):
        self.message = message
        self.process = process
        self.event = event
        self.proc_errors = proc_errors
        super().__init__(message)

    @property
    def formatted_errors(self):
        return [{
            'id': str(pe.id),
            'func': str(pe.func),
            'message': str(pe.message)
        } for pe, _ in self.proc_errors]



class Price:
    def __init__(self, amount, currency):
        self.amount = amount
        self.currency = currency

    def __str__(self):
        return f'{self.currency} {self.amount}'

    def __eq__(self, other):
        return self.amount == other.amount and self.currency == other.currency



import json
import logging
from .models import (Account, Card, Charge, Coupon, Customer, Event, Invoice,
                     InvoiceItem, Order, Plan, Product, SKU, Source, Subscription,
                     Transfer, Refund, BankAccount, Payout, Application, IssuerFraudRecord,
                     RayureMeta, Dispute, RayureEventProcess, RayureEventProcessingError,
                     BalanceTransaction)
from django.contrib import admin
from django.contrib.postgres.fields import JSONField
from django.forms import widgets
from django.utils.html import format_html
from django.urls import reverse

logger = logging.getLogger('rayures')


class HasCustomerFilter(admin.SimpleListFilter):
    title = 'customer'
    parameter_name = 'has_customer'

    def lookups(self, request, model_admin):
        return [
            ('yes', 'Has customer'),
            ('no', 'No customer')
        ]

    def queryset(self, request, queryset):
        if self.value() == 'yes':
            return queryset.filter(customer__isnull=False)
        if self.value() == 'no':
            return queryset.filter(customer__isnull=True)
        return queryset


class PrettyJsonWidget(widgets.Textarea):
    def __init__(self, attrs=None):
        attrs = attrs or {}
        attrs.setdefault('cols', 80)
        attrs.setdefault('rows', 30)
        super().__init__(attrs)

    def format_value(self, value):
        value = json.loads(value)
        value = json.dumps(value, indent=4)
        return super().format_value(value)


class ModelAdmin(admin.ModelAdmin):
    formfield_overrides = {
        JSONField: {'widget': PrettyJsonWidget}
    }

    def has_add_permission(self, request):
        return False

    def get_list_display(self, request):
        list_display = super().get_list_display(request)
        if 'deleted_at' not in list_display:
            list_display = list_display + ('deleted_at',)
        if 'show_events_url' not in list_display:
            list_display = list_display + ('show_events_url',)
        return list_display

    def get_list_filter(self, request):
        """
        Return a sequence containing the fields to be displayed as filters in
        the right sidebar of the changelist page.
        """
        list_filter = super().get_list_filter(request)
        for field in self.model._meta.get_fields():
            if field.name in ('status', 'type', 'created_at', 'livemode') and field.name not in list_filter:
                list_filter = list_filter + (field.name,)
        if 'deleted_at' not in list_filter:
            list_filter = list_filter + ('deleted_at',)
        return list_filter

    def show_events_url(self, obj):
        url = f'{reverse("admin:rayures_event_changelist")}?q={obj.id}'
        return format_html(f'<a href="{url}">show events</a>')
    show_events_url.allow_tags = True
    show_events_url.short_description = 'Events'


@admin.register(RayureMeta)
class RayureMetaAdmin(admin.ModelAdmin):
    list_display = 'id', 'content_type', 'created_at', 'updated_at', 'deleted_at', 'show_obj_url', 'source'
    search_field = '=id', 'created_at', 'updated_at', 'deleted_at', '=event_id', '=request_id'
    readonly_fields = ('id', 'content_type', 'created_at', 'updated_at',
                       'deleted_at', 'event', 'request_id', 'idempotency_key', 'source',)
    list_filter = 'content_type', 'created_at', 'updated_at', 'deleted_at', 'source'

    def has_add_permission(self, request):
        return False

    def show_obj_url(self, obj):
        if obj.id is not None:
            app_label, model = obj.content_type.app_label, obj.content_type.model
            url = f'{reverse(f"admin:{app_label}_{model}_changelist")}?q={obj.id}'
            return format_html(f'<a href="{url}">show object</a>')
    show_obj_url.allow_tags = True
    show_obj_url.short_description = 'Object'


@admin.register(RayureEventProcess)
class RayureEventProcessAdmin(admin.ModelAdmin):
    list_display = 'id', 'event', 'status', 'started_at', 'ended_at', 'show_event_url'
    search_field = '=id', '=event', '=errors__id', 'status', 'started_at', 'ended_at',
    readonly_fields = 'id', 'event', 'status', 'started_at', 'ended_at',

    def has_add_permission(self, request):
        return False

    def show_event_url(self, obj):
        url = f'{reverse("admin:rayures_event_changelist")}?q={obj.event_id}'
        return format_html(f'<a href="{url}">show event</a>')
    show_event_url.allow_tags = True
    show_event_url.short_description = 'Events'


@admin.register(RayureEventProcessingError)
class RayureEventProcessingErrorAdmin(admin.ModelAdmin):
    list_display = 'id', 'created_at', 'message', 'process', 'func',
    search_field = '=id', '=process', '=process__event',
    readonly_fields = 'id', 'process', 'created_at', 'acknowledged_at', 'func', 'message', 'traceback', 'http_body',

    def has_add_permission(self, request):
        return False


@admin.register(BalanceTransaction)
class BalanceTransactionAdmin(ModelAdmin):
    list_display = 'id', 'amount', 'available_on', 'created_at', 'source', 'status', 'type',
    search_field = '=id', 'available_on', 'created_at', 'status', 'type',
    list_filter = 'status', 'type', 'source', 'available_on', 'created_at',


@admin.register(Dispute)
class DisputeAdmin(ModelAdmin):
    list_display = 'id', 'amount', 'balance_transaction', 'charge', 'created_at', 'reason', 'status'
    search_field = '=id', '=balance_transaction', '=charge',
    list_filter = 'status', 'reason', 'created_at',


@admin.register(Refund)
class RefundAdmin(ModelAdmin):
    list_display = 'id', 'receipt_number', 'status', 'balance_transaction', 'charge'
    search_field = '=id', '=balance_transaction', '=charge',


@admin.register(IssuerFraudRecord)
class IssuerFraudRecordAdmin(ModelAdmin):
    list_display = 'id', 'charge', 'created_at', 'post_date', 'fraud_type'
    search_field = '=id', '=charge', 'created_at', 'post_date',
    readonly_fields = 'charge',


@admin.register(Payout)
class PayoutAdmin(ModelAdmin):
    list_display = 'id', 'status', 'type', 'balance_transaction', 'amount', 'created_at', 'destination', 'failure_code'
    list_filter = 'status', 'type', 'created_at',
    search_field = '=id', '=balance_transaction', '=destination'


@admin.register(Account)
class AccountAdmin(ModelAdmin):
    list_display = 'id', 'type', 'support_email', 'support_phone',
    search_fields = '=id',


@admin.register(Application)
class ApplicationAdmin(ModelAdmin):
    list_display = 'id', 'name',
    search_fields = '=id',


@admin.register(BankAccount)
class BankAccountAdmin(ModelAdmin):
    list_display = 'id', 'customer', 'status'
    list_filter = 'status',
    search_fields = '=id',
    readonly_fields = 'customer',


@admin.register(Coupon)
class CouponAdmin(ModelAdmin):
    list_display = 'id', 'amount_off', 'created_at', 'valid', 'duration'
    list_filter = 'valid', 'created_at',
    search_fields = '=id',


@admin.register(Invoice)
class InvoiceAdmin(ModelAdmin):
    list_display = ('id', 'total', 'amount_due', 'amount_paid', 'amount_remaining',
                    'period_start_at', 'period_end_at', 'date', 'due_date', 'paid',
                    'forgiven', 'attempted', 'closed', 'total', 'subtotal',
                    'charge_id', 'customer_id', 'subscription_id',
                    'webhooks_delivered_at', 'number')
    list_filter = 'paid', 'date', 'due_date', 'forgiven', 'closed'
    search_fields = '=id', '=charge_id', '=customer_id', '=subscription_id'
    readonly_fields = 'charge', 'subscription', 'customer',


@admin.register(InvoiceItem)
class InvoiceItemAdmin(ModelAdmin):
    list_display = ('id', 'date', 'amount', 'customer_id', 'plan_id',
                    'subscription_id', 'invoice_id', 'amount', 'quantity',
                    'period_start_at', 'period_end_at')
    list_filter = 'date',
    search_fields = '=id', '=plan_id', '=customer_id', '=subscription_id'
    readonly_fields = 'plan', 'subscription', 'invoice', 'customer',


@admin.register(Order)
class OrderAdmin(ModelAdmin):
    list_display = ('id', 'amount', 'amount_returned', 'application_fee', 'email',
                    'charge_id', 'customer_id', 'status', 'created_at', 'updated_at',
                    'paid_at', 'canceled_at', 'fulfiled_at', 'returned_at')
    list_filter = 'status', 'created_at', 'updated_at', 'paid_at', 'canceled_at', 'fulfiled_at', 'returned_at',
    search_fields = '=id', '=charge_id', '=customer_id'


@admin.register(Product)
class ProductAdmin(ModelAdmin):
    list_display = 'id', 'name', 'url', 'type', 'active', 'created_at', 'updated_at'
    list_filter = 'active', 'type', 'created_at', 'updated_at'
    search_fields = '=id',


@admin.register(SKU)
class SKUAdmin(ModelAdmin):
    list_display = 'id', 'price', 'active', 'updated_at', 'created_at', 'product_id'
    list_filter = 'active', 'created_at', 'updated_at'
    search_fields = '=id',


@admin.register(Card)
class CardAdmin(ModelAdmin):
    list_display = 'id', 'name', 'brand', 'customer_id', 'last4', 'exp_year', 'exp_month'
    list_filter = 'brand',
    search_fields = '=id', '=customer_id'
    readonly_fields = 'customer',


@admin.register(Source)
class SourceAdmin(ModelAdmin):
    list_display = 'id', 'amount', 'created_at', 'type', 'usage', 'status', 'customer_id',
    list_filter = 'status', 'type', 'created_at', HasCustomerFilter,
    search_fields = '=id', '=customer_id',


@admin.register(Transfer)
class TransferAdmin(ModelAdmin):
    list_display = ('id', 'amount', 'arrival_date', 'date', 'type', 'method',
                    'status', 'created_at', 'balance_transaction_id')
    list_filter = 'type', 'method', 'status'
    search_fields = '=id', '=balance_transaction_id'


@admin.register(Charge)
class ChargeAdmin(ModelAdmin):
    list_display = ('id', 'status', 'invoice_id', 'customer_id', 'paid', 'order_id',
                    'source_id', 'invoice_id', 'amount', 'amount_refunded',
                    'balance_transaction_id')
    list_filter = 'paid', 'created_at', 'status',
    search_fields = '=id', '=invoice_id', '=customer_id', '=order_id', '=source_id', '=invoice_id'
    readonly_fields = 'customer', 'invoice', 'order', 'source', 'balance_transaction', 'created_at'


@admin.register(Customer)
class CustomerAdmin(ModelAdmin):
    list_display = ('id', 'email', 'delinquent', 'invoice_prefix', 'created_at',
                    'account_balance', 'default_source_display',)
    list_filter = 'created_at', 'delinquent', 'deleted_at',
    search_fields = '=id', '=email'
    readonly_fields = 'default_source',

    def default_source_display(self, obj):
        src = obj.default_source
        if src:
            return src.id
    default_source_display.allow_tags = True
    default_source_display.short_description = 'Default source'


@admin.register(Subscription)
class SubscriptionAdmin(ModelAdmin):
    list_display = ('id', 'start_at', 'created_at', 'ended_at', 'trial_start_at',
                    'trial_end_at', 'canceled_at', 'current_period_end_at',
                    'current_period_start_at', 'billing_cycle_anchor', 'status',
                    'billing', 'customer_id', 'plan_id', 'cancel_at_period_end',
                    'quantity', 'days_until_due')
    list_filter = 'plan_id', 'status', 'created_at', 'ended_at'
    search_fields = '=id', '=customer_id', '=plan_id'


@admin.register(Plan)
class PlanAdmin(ModelAdmin):
    list_display = 'id', 'name', 'interval', 'interval_count', 'trial_period_days', 'created_at'
    list_filter = 'interval', 'trial_period_days'
    readonly_fields = 'id', 'api_version', 'name', 'interval', 'interval_count', 'trial_period_days', 'created_at'
    search_fields = '=id',


@admin.register(Event)
class EventAdmin(ModelAdmin):
    list_display = 'id', 'type', 'request_id', 'idempotency_key', 'created_at', 'object_id', 'show_obj_url',
    list_filter = 'type', 'created_at',
    readonly_fields = 'id', 'api_version', 'type', 'created_at', 'pending_webhooks', 'content_type', 'object_id',
    search_fields = '=id', '=request_id', '=object_id'

    def get_list_display(self, request):
        list_display = super().get_list_display(request)
        list_display = tuple(attr for attr in list_display if attr != 'show_events_url')
        if 'show_obj_url' not in list_display:
            list_display = list_display + ('show_obj_url',)
        return list_display

    def show_obj_url(self, obj):
        if obj.object_id is not None:
            try:
                app_label, model = obj.content_type.app_label, obj.content_type.model
                url = f'{reverse(f"admin:{app_label}_{model}_changelist")}?q={obj.object_id}'
                return format_html(f'<a href="{url}">show object</a>')
            except Exception as error:
                logger.warn(f'{app_label}.{model} has no admin entry')
                return '-'
    show_obj_url.allow_tags = True
    show_obj_url.short_description = 'Object'



import logging
import stripe
import traceback
from . import fields
from .utils import price_from_stripe, dt_from_stripe
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import JSONField
from django.db import models
from django.db.models.base import ModelBase
from django.db.models.functions import Coalesce, Now

logger = logging.getLogger('rayures')
PERSISTED_MODELS = {}
STRIPE_OBJECTS = {}


def state_factory(instance):
    if isinstance(instance, dict):
        assert 'object' in instance
        assert 'id' in instance
        data = instance
        version = stripe.api_version
    elif isinstance(instance, (Entity, PersistedModel)):
        data = instance.data
        version = getattr(instance, 'api_version', None)
    else:
        raise ValueError('will not work man')
    return stripe.convert_to_stripe_object(data,
                                           api_key=stripe.api_key,
                                           stripe_account=None,
                                           stripe_version=version)


class SoftDeletionQuerySet(models.QuerySet):
    # TODO: add an attribute "persisted: True everytime a row is instantiated"

    def delete(self):
        return super().update(deleted_at=Coalesce(models.F('deleted_at'), Now()))

    def hard_delete(self):
        return super().delete()

    def alive(self):
        return self.filter(deleted_at=None)

    def dead(self):
        return self.exclude(deleted_at=None)


class StripeMeta(type):
    def __new__(metacls, name, bases, namespace, object=None):
        global STRIPE_OBJECTS
        cls = super().__new__(metacls, name, bases, namespace)
        if object:
            STRIPE_OBJECTS[object] = cls
            cls._stripe_object = object
        return cls


class PersistedMeta(ModelBase):
    def __new__(metacls, name, bases, namespace, object=None):
        global PERSISTED_MODELS
        cls = super().__new__(metacls, name, bases, namespace)
        if object:
            STRIPE_OBJECTS[object] = cls
            PERSISTED_MODELS[object] = cls
            cls._stripe_object = object
        return cls


class PersistedModel(models.Model, metaclass=PersistedMeta):
    id = models.CharField(max_length=255, primary_key=True, editable=False)
    api_version = models.CharField(max_length=12, editable=False)
    data = JSONField(default=dict)

    deleted_at = models.DateTimeField(editable=False, null=True)
    objects = SoftDeletionQuerySet.as_manager()

    @classmethod
    def from_db(cls, db, field_names, values):
        # Append persisted and deleted hints
        instance = super().from_db(db, field_names, values)
        instance.persisted = True
        instance.deleted = False
        return instance

    @property
    def events(self):
        # got events without the mess of GenericRelation('rayures.Event')
        return Event.objects.filter(object_id=self.id)

    def delete(self):
        self.deleted_at = self.deleted_at or Now()
        self.save(update_fields=['deleted_at'])

    def hard_delete(self):
        super().delete()

    def rebound_fields(self):
        for field in self._meta.private_fields:
            if isinstance(field, fields.StripeField):
                field.rebound(self)
        return self

    def refresh_from_stripe(self):
        """Refresh current object from stripe.
        """
        if self.deleted_at is not None:
            # TODO: warn won't refresh because it's deleted
            return
        state = state_factory(self)
        try:
            state.refresh()
        except stripe.error.InvalidRequestError as err:
            # TODO: mark it deleted now, but should have been done previously
            logger.warn(f"Tried to refresh {self.id} from stripe but seems to not exists: {err}")
            self.delete()
        else:
            self.refresh_from_state(state)

    def refresh_from_state(self, state):
        """Refresh from a stripe state
        """
        assert self.id == state['id']
        if state.get('deleted', False):
            return self.delete()
        assert self.data['object'] == state['object']
        self.data = state
        self.save(update_fields=['data'])
        self.rebound_fields()

    def refresh_related_objects(self, *names, persist=False):
        """Refresh related objects
        """
        fields = [
            f for f in self._meta.get_fields()
            if isinstance(f, fields.ForeignKey)
        ]
        if names:
            fields = [f for f in fields if f.name in names]

        fom = ((f, getattr(f.related_model, '_stripe_object'), f.related_model) for f in fields)
        fom = [(f, o, m) for f, o, m in fom if o is not None]

        results = {}
        for f, o, m in fom:
            stripe_id = getattr(self, f'{f.name}_id', None)
            if stripe_id is None:
                results[f.name] = None
            else:
                state = state_factory({'object': o, 'id': stripe_id})
                state.refresh()
                results[f.name], _ = m.ingest(state, api_version=stripe.api_version, persist=persist)
        return results

    class Meta:
        abstract = True

    def __repr__(self):
        return f'<{type(self)}({self.id})>'

    def __str__(self):
        return str(self.id or '-')


class Entity(metaclass=StripeMeta):
    def __init__(self, data):
        self.data = data

    @property
    def id(self):
        return self.data['id']

    @property
    def events(self):
        # got events without the mess of GenericRelation('rayures.Event')
        return Event.objects.filter(object_id=self.data['id'])

    def refresh_from_stripe(self):
        """Refresh current object from stripe.
        """
        state = state_factory(self)
        state.refresh()
        self.refresh_from_state(state)

    def refresh_from_state(self, state):
        """Refresh from a stripe state
        """
        assert self.id == state['id']
        self.data = state

    def __repr__(self):
        return f'<{type(self)}({self.id})>'

    def __str__(self):
        return str(self.id or '-')


class Account(PersistedModel, object='account'):
    country = fields.CharField(source='country')
    created_at = fields.DateTimeField(source='created')
    debit_negative_balances = fields.BooleanField(source='debit_negative_balances')
    email = fields.CharField(source='email')
    support_email = fields.CharField(source='support_email')
    support_phone = fields.CharField(source='support_phone')
    type = fields.CharField(source='type')


class Application(PersistedModel, object='application'):
    """Custom object, not documented in Stripe as for 2018-06-19
    """
    name = fields.CharField(source='name')


class IssuerFraudRecord(PersistedModel, object='issuer_fraud_record'):
    """Custom object, not documented in Stripe as for 2018-06-19
    """
    charge = fields.ForeignKey('rayures.Charge', related_name='issuer_fraud_records', source='charge')
    created_at = fields.DateTimeField(source='created')
    post_date = fields.DateTimeField(source='created')
    fraud_type = fields.CharField(source='fraud_type')


class Payout(PersistedModel, object='payout'):
    type = fields.CharField(source='type')
    amount = fields.PriceField(source='amount')
    method = fields.CharField(source='method')
    status = fields.CharField(source='status')
    created_at = fields.DateTimeField(source='created')
    source_type = fields.CharField(source='source_type')
    arrival_date = fields.DateTimeField(source='created')
    failure_code = fields.CharField(source='failure_code')
    failure_message = fields.CharField(source='failure_message')
    balance_transaction = fields.ForeignKey('rayures.BalanceTransaction',
                                            related_name='+',
                                            source='balance_transaction')
    failure_balance_transaction = fields.ForeignKey('rayures.BalanceTransaction',
                                                    related_name='+',
                                                    source='failure_balance_transaction')
    destination = fields.ForeignKey('rayures.BankAccount',
                                    related_name='+',
                                    source='destination')
    metadata = fields.HashField(source='metadata')

    @property
    def failure(self):
        return PayoutFailure(self.failure_code, self.failure_message, self.failure_balance_transaction)


class BankAccount(PersistedModel, object='bank_account'):
    account_holder_name = fields.CharField(source='account_holder_name')
    account_holder_type = fields.CharField(source='account_holder_type')
    bank_name = fields.CharField(source='bank_name')
    currency = fields.CharField(source='currency')
    fingerprint = fields.CharField(source='fingerprint')
    last4 = fields.CharField(source='last4')
    routing_number = fields.CharField(source='routing_number')
    status = fields.CharField(source='status')
    customer = fields.ForeignKey('rayures.Customer', related_name='+', source='customer')
    default_for_currency = fields.BooleanField(source='default_for_currency')
    metadata = fields.HashField(source='metadata')


class BalanceTransaction(PersistedModel, object='balance_transaction'):
    amount = fields.PriceField(source='amount')
    available_on = fields.DateTimeField(source='available_on')
    created_at = fields.DateTimeField(source='created')
    fee = fields.PriceField(source='amount')
    net = fields.PriceField(source='amount')
    source = fields.ForeignKey('rayures.BalanceTransaction', related_name='+', source='source')
    status = fields.CharField(source='status')
    type = fields.CharField(source='type')


class Charge(PersistedModel, object='charge'):
    amount = fields.PriceField(source='amount')
    amount_refunded = fields.PriceField(source='amount_refunded')
    balance_transaction = fields.ForeignKey('rayures.BalanceTransaction',
                                            related_name='charges',
                                            source='balance_transaction')
    captured = fields.BooleanField(source='captured')
    created_at = fields.DateTimeField(source='created')
    # customer_id = fields.CharField(source='customer')
    customer = fields.ForeignKey('rayures.Customer', related_name='charges', source='customer')

    status = fields.CharField(source='status')

    # invoice_id = fields.CharField(source='invoice')
    invoice = fields.ForeignKey('rayures.Invoice', related_name='charges', source='invoice')
    paid = fields.BooleanField(source='paid')
    refunded = fields.BooleanField(source='refunded')
    # order_id = fields.CharField(source='order')
    order = fields.ForeignKey('rayures.Order', related_name='charges', source='order')
    # source_id = fields.CharField(source='source.id')
    source = fields.ForeignKey('rayures.Source', related_name='charges', source='source.id')
    balance_transaction = fields.ForeignKey('rayures.BalanceTransaction',
                                            related_name='charges',
                                            source='balance_transaction')
    metadata = fields.HashField(source='metadata')


class Card(PersistedModel, object='card'):
    name = fields.CharField(source='name')
    brand = fields.CharField(source='brand')
    last4 = fields.CharField(source='last4')
    # customer_id = fields.CharField(source='customer')
    customer = fields.ForeignKey('rayures.Customer', related_name='+', source='customer')
    exp_year = fields.IntegerField(source='exp_year')
    exp_month = fields.IntegerField(source='exp_month')
    fingerprint = fields.CharField(source='fingerprint')
    funding = fields.CharField(source='funding')
    available_payout_methods = fields.CharField(source='available_payout_methods')
    cvc_check = fields.CharField(source='cvc_check')
    metadata = fields.HashField(source='metadata')


class Coupon(PersistedModel, object='coupon'):
    amount_off = fields.PriceField(source='amount_off')
    percent_off = fields.IntegerField(source='percent_off')
    created_at = fields.DateTimeField(source='created')
    valid = fields.BooleanField(source='valid')
    duration = fields.CharField(source='duration')
    name = fields.CharField(source='name')
    redeem_by = fields.DateTimeField(source='redeem_by')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')


class Invoice(PersistedModel, object='invoice'):
    total = fields.PriceField(source='total')
    starting_balance = fields.PriceField(source='starting_balance')
    ending_balance = fields.PriceField(source='ending_balance')
    invoice_pdf = fields.CharField(source='invoice_pdf')
    amount_due = fields.PriceField(source='amount_due')
    amount_paid = fields.PriceField(source='amount_paid')
    amount_remaining = fields.PriceField(source='amount_remaining')
    period_start_at = fields.DateTimeField(source='period_start')
    period_end_at = fields.DateTimeField(source='period_end')
    date = fields.DateTimeField(source='date')
    due_date = fields.DateTimeField(source='due_date')
    next_payment_attempt = fields.DateTimeField(source='next_payment_attempt')
    paid = fields.BooleanField(source='paid')
    forgiven = fields.BooleanField(source='forgiven')
    attempted = fields.BooleanField(source='attempted')
    closed = fields.BooleanField(source='closed')
    total = fields.PriceField(source='total')
    subtotal = fields.PriceField(source='subtotal')
    # charge_id = fields.CharField(source='charge')
    charge = fields.ForeignKey('rayures.Charge', related_name='invoices', source='charge')
    # customer_id = fields.CharField(source='customer')
    customer = fields.ForeignKey('rayures.Customer', related_name='invoices', source='customer')
    # subscription_id = fields.CharField(source='subscription')
    subscription = fields.ForeignKey('rayures.Subscription', related_name='invoices', source='subscription')
    webhooks_delivered_at = fields.DateTimeField(source='webhooks_delivered_at')
    hosted_invoice_url = fields.CharField(source='hosted_invoice_url')
    number = fields.CharField(source='number')
    receipt_number = fields.CharField(source='receipt_number')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')

    @property
    def discount(self):
        if self.data['discount']:
            return Discount(self.data['discount'])


class InvoiceItem(PersistedModel, object='invoiceitem'):
    date = fields.DateTimeField(source='date')
    amount = fields.PriceField(source='amount')
    plan = fields.ForeignKey('rayures.Plan', related_name='invoice_items', source='plan')
    subscription = fields.ForeignKey('rayures.Subscription', related_name='invoice_items', source='subscription')
    invoice = fields.ForeignKey('rayures.Invoice', related_name='invoice_items', source='invoice')
    customer = fields.ForeignKey('rayures.Customer', related_name='invoice_items', source='customer')
    period_start_at = fields.DateTimeField(source='period.start')
    period_end_at = fields.DateTimeField(source='period.end')
    quantity = fields.IntegerField(source='quantity')
    proration = fields.BooleanField(source='proration')
    discountable = fields.BooleanField(source='discountable')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')


class Order(PersistedModel, object='order'):
    amount = fields.PriceField(source='amount')
    amount_returned = fields.PriceField(source='amount_returned')
    application_fee = fields.PriceField(source='application_fee')
    email = fields.CharField(source='email')
    charge = fields.ForeignKey('rayures.Charge', related_name='orders', source='charge')
    customer = fields.ForeignKey('rayures.Customer', related_name='orders', source='customer')
    status = fields.CharField(source='status')
    created_at = fields.DateTimeField(source='created')
    updated_at = fields.DateTimeField(source='updated')
    paid_at = fields.DateTimeField(source='status_transitions.paid')
    canceled_at = fields.DateTimeField(source='status_transitions.canceled')
    fulfiled_at = fields.DateTimeField(source='status_transitions.fulfiled')
    returned_at = fields.DateTimeField(source='status_transitions.returned')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')


class Product(PersistedModel, object='product'):
    name = fields.CharField(source='name')
    url = fields.CharField(source='url')
    type = fields.CharField(source='type')
    caption = fields.CharField(source='caption')
    active = fields.BooleanField(source='active')
    created_at = fields.DateTimeField(source='created')
    updated_at = fields.DateTimeField(source='updated')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')


class Refund(PersistedModel, object='refund'):
    amount = fields.PriceField(source='amount')
    balance_transaction = fields.ForeignKey('rayures.BalanceTransaction',
                                            related_name='charges',
                                            source='balance_transaction')
    charge = fields.ForeignKey('rayures.Charge', related_name='refunds', source='charge')
    created_at = fields.DateTimeField(source='created')
    reason = fields.CharField(source='reason')
    receipt_number = fields.CharField(source='receipt_number')
    status = fields.CharField(source='status')

    name = fields.CharField(source='name')
    url = fields.CharField(source='url')
    type = fields.CharField(source='type')
    caption = fields.CharField(source='caption')
    active = fields.BooleanField(source='active')
    created_at = fields.DateTimeField(source='created')
    updated_at = fields.DateTimeField(source='updated')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')


# TODO: Token ?

class SKU(PersistedModel, object='sku'):
    price = fields.PriceField(source='price')
    active = fields.BooleanField(source='active')
    created_at = fields.DateTimeField(source='created')
    updated_at = fields.DateTimeField(source='updated')
    product = fields.ForeignKey('rayures.Product', related_name='skus', source='product')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')


class Source(PersistedModel, object='source'):
    amount = fields.PriceField(source='amount')
    created_at = fields.DateTimeField(source='created')
    type = fields.CharField(source='type')
    usage = fields.CharField(source='usage')
    status = fields.CharField(source='status')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')
    customer = fields.ForeignKey('rayures.Customer', related_name='+', source='customer')

    @property
    def card(self):
        card = getattr(self.data, 'card', None)
        if card:
            return SourceCard(card)


class SourceCard:
    def __init__(self, data):
        self.data = data

    @property
    def brand(self):
        return self.data['brand']

    @property
    def last4(self):
        return self.data['last4']

    @property
    def exp_year(self):
        return self.data['exp_year']

    @property
    def exp_month(self):
        return self.data['exp_month']

    @property
    def fingerprint(self):
        return self.data['fingerprint']

    @property
    def cvc_check(self):
        return self.data['cvc_check']


class Transfer(PersistedModel, object='transfer'):
    amount = fields.PriceField(source='amount')
    arrival_date = fields.DateTimeField(source='arrival_date')
    date = fields.DateTimeField(source='date')
    type = fields.CharField(source='type')
    method = fields.CharField(source='method')
    status = fields.CharField(source='status')
    created_at = fields.DateTimeField(source='created')
    balance_transaction_id = fields.CharField(source='balance_transaction')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')


class Customer(PersistedModel, object='customer'):
    email = fields.CharField(source='email')
    invoice_prefix = fields.CharField(source='invoice_prefix')
    created_at = fields.DateTimeField(source='created')
    account_balance = fields.PriceField(source='account_balance')
    # default_source_id = fields.PriceField(source='default_source')
    # default_source = fields.ForeignKey('rayures.Source', related_name='+', source='default_source')
    delinquent = fields.BooleanField(source='delinquent')
    # discount_id = fields.CharField(source='discount')
    # discount = fields.ForeignKey('rayures.Source', related_name='customers', source='discount')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')

    @property
    def default_source(self):
        """Returns default source.

        Object can be either a Card or a Source.
        """
        # TODO: make it a relation field searchable by source/card id
        if self.data['default_source']:
            source = self.data['default_source']
            if isinstance(source, str):
                for s in self.data['sources']['data']:
                    # check in local data
                    if s['id'] == source:
                        source = s
                        break
                    else:
                        # worst case, must performs a query
                        obj = Source.objects.filter(id=source).first()
                        if not obj:
                            obj = Card.objects.filter(id=source).first()
                        return obj
            else:
                # must be an embedded object
                assert 'id' in source
                assert 'object' in source
            if source:
                from .reconciliation import reconciliate
                return reconciliate(source, persist=False, api_version=self.api_version, source="orm").instance

    @property
    def sources(self):
        from .reconciliation import reconciliate
        sources = []
        for s in self.data['sources']['data']:
            source = reconciliate(s, persist=False, api_version=self.api_version, source="orm").instance
            sources.append(source)
        if self.data['default_source']:
            ds = self.data['default_source']
            if isinstance(ds, dict):
                ds = ds['id']
                sources = sorted(sources, key=lambda x: 0 if x.id == ds else 1)
        return sources

    @property
    def discount(self):
        if self.data['discount']:
            return Discount(self.data['discount'])


class Dispute(PersistedModel, object='dispute'):
    amount = fields.PriceField(source='amount')
    balance_transaction = fields.ForeignKey('rayures.BalanceTransaction',
                                            related_name='disputes',
                                            source='balance_transaction')
    # FIXME: stripe expose a balance_transactions array[]. see how to expose it?
    # charge_id = fields.CharField(source='charge')
    charge = fields.ForeignKey('rayures.Charge', related_name='disputes', source='charge')
    created_at = fields.DateTimeField(source='created')
    reason = fields.CharField(source='reason')
    status = fields.CharField(source='status')
    livemode = fields.BooleanField(source='livemode')
    metadata = fields.HashField(source='metadata')


class Subscription(PersistedModel, object='subscription'):
    start_at = fields.DateTimeField(source='start')
    created_at = fields.DateTimeField(source='created')
    ended_at = fields.DateTimeField(source='ended_at')
    trial_start_at = fields.DateTimeField(source='trial_start')
    trial_end_at = fields.DateTimeField(source='trial_end')
    canceled_at = fields.DateTimeField(source='canceled_at')
    current_period_end_at = fields.DateTimeField(source='current_period_end')
    current_period_start_at = fields.DateTimeField(source='current_period_start')
    billing_cycle_anchor = fields.DateTimeField(source='billing_cycle_anchor')
    status = fields.CharField(source='status')
    billing = fields.CharField(source='billing')
    # customer_id = fields.CharField(source='customer')
    customer = fields.ForeignKey('rayures.Customer', related_name='subscriptions', source='customer')
    # plan_id = fields.CharField(source='plan.id')
    plan = fields.ForeignKey('rayures.Plan', related_name='subscriptions', source='plan.id')
    cancel_at_period_end = fields.BooleanField(source='cancel_at_period_end')
    # discount_id = fields.CharField(source='discount')
    # discount = fields.ForeignKey('rayures.Discount', related_name='subscriptions', source='discount')
    quantity = fields.IntegerField(source='quantity')
    days_until_due = fields.IntegerField(source='days_until_due')
    livemode = fields.BooleanField(source='livemode')
    # TODO: items
    metadata = fields.HashField(source='metadata')

    @property
    def discount(self):
        if self.data['discount']:
            return Discount(self.data['discount'])

    @property
    def items(self):
        # TODO: generic nested items handler
        # TODO: handle has_more ?
        for data in self.data['items']['data']:
            yield SubscriptionItem(data)


class SubscriptionItem(Entity, object='subcription_item'):
    @property
    def created_at(self):
        return dt_from_stripe(self.data['created_at'])

    @property
    def metadata(self):
        return self.data['metadata']

    @property
    def plan_id(self):
        return self.data['plan']['id']

    @property
    def quantity(self):
        return self.data['quantity']

    @property
    def subscription_id(self):
        return self.data['subscription']


class Plan(PersistedModel, object='plan'):
    name = fields.CharField(source='name')
    interval = fields.CharField(source='interval')
    interval_count = fields.IntegerField(source='interval_count')
    trial_period_days = fields.IntegerField(source='trial_period_days')
    created_at = fields.DateTimeField(source='created')
    amount = fields.PriceField(source='amount')
    aggregate_usage = fields.CharField(source='aggregate_usage')
    billing_scheme = fields.CharField(source='billing_scheme')
    usage_type = fields.CharField(source='usage_type')
    livemode = fields.BooleanField(source='livemode')
    product = fields.ForeignKey('rayures.Product', related_name='plans', source='product')
    metadata = fields.HashField(source='metadata')
    active = fields.BooleanField(source='active')
    billing_scheme = fields.CharField(source='billing_scheme')
    # ? tiers
    # ? tiers_mode
    # ? transform_usage
    usage_type = fields.CharField(source='usage_type')

    # statement_descriptor in product
    # name in product


class Event(PersistedModel, object='event'):
    pending_webhooks = fields.IntegerField(source='pending_webhooks')
    type = fields.CharField(source='type')
    request_id = fields.CharField(source='request.id')
    idempotency_key = fields.CharField(source='request.idempotency_key')
    created_at = fields.DateTimeField(source='created')
    api_version = fields.CharField(source='api_version')
    livemode = fields.BooleanField(source='livemode')

    # obj
    content_type = models.ForeignKey(ContentType, on_delete=models.SET_DEFAULT, null=True, default=None)
    object_id = models.CharField(default=None, max_length=50, null=True)
    content_object = GenericForeignKey('content_type', 'object_id')


class Discount(Entity, object="discount"):

    @property
    def end_at(self):
        return dt_from_stripe(self.data['end'])

    @property
    def start_at(self):
        return dt_from_stripe(self.data['start'])

    @property
    def coupon(self):
        if self['coupon']:
            return Coupon.objects.get(id=self['coupon']['id'])

    @property
    def customer(self):
        if self['customer']:
            return Customer.objecst.get(id=self['customer']['id'])

    @property
    def subscription(self):
        if self['subscription']:
            return Subscription.objecst.get(id=self['subscription']['id'])


class Balance(Entity, object='balance'):
    @property
    def available(self):
        for item in self.data['available']:
            yield BalanceItem(item)

    @property
    def pending(self):
        for item in self.data['pending']:
            yield BalanceItem(item)


class BalanceItem(Entity):
    @property
    def amount(self):
        return price_from_stripe(self.data['amount'], self.data['currency'])

    @property
    def source_types(self):
        return self.data['source_types']


class PayoutFailure:
    def __init__(self, code, message, balance_transaction):
        self.code = code
        self.message = message
        self.balance_transaction = balance_transaction


# TODO: order_return

class OrderReturn(PersistedModel, object='order_return'):
    amount = fields.PriceField(source='amount')
    created_at = fields.DateTimeField(source='created')
    # todo: items
    # description = fields.CharField(source='description')
    # parent = fields.CharField(source='parent')
    # quantity = fields.IntegerField(source='quantity')
    livemode = fields.BooleanField(source='livemode')
    order = fields.ForeignKey('rayures.Order', related_name='order_returns', source='order')
    refund = fields.ForeignKey('rayures.Refund', related_name='order_returns', source='refund')


class Trace:
    def __init__(self, func, subcalls, parent):
        self.data = {
            'func': f'{func.__module__}.{func.__qualname__}',
            'api_calls': subcalls
        }
        self.parent = parent

    def log_error(self, error):
        # most of stripe.StripeError have http_body
        http_body = getattr(error, 'http_body', None) or ''
        lines = traceback.format_exception(type(error), error, error.__traceback__)
        formatted = ''.join(lines)

        return self.parent.errors.create(
            http_body=http_body,
            func=self.data['func'],
            message=str(error),
            traceback=formatted)


class RayureEventProcess(models.Model):
    STATUS_CHOICES = [
        ('received', 'received'),
        ('success', 'success'),
        ('failure', 'failure')
    ]
    event = models.ForeignKey("rayures.Event", related_name='+', on_delete=models.CASCADE)
    status = models.CharField(choices=STATUS_CHOICES, max_length=12, default='received')
    started_at = models.DateTimeField(auto_now_add=True)
    ended_at = models.DateTimeField(editable=False, null=True)
    traces = JSONField(default=list)

    def log_trace(self, func, subcalls):
        trace = Trace(func, subcalls, self)
        self.traces.append(trace.data)
        return trace

    class Meta:
        verbose_name = 'Rayure event process'
        verbose_name_plural = 'Rayure event processes'


class RayureEventProcessingError(models.Model):
    process = models.ForeignKey("rayures.RayureEventProcess",
                                related_name='errors',
                                on_delete=models.CASCADE)
    func = models.TextField(null=True)
    http_body = models.TextField()
    message = models.TextField()
    traceback = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True)
    acknowledged_at = models.DateTimeField(null=True)


class RayureMeta(models.Model):
    id = models.CharField(max_length=255, primary_key=True)
    content_type = models.ForeignKey(ContentType,
                                     on_delete=models.CASCADE,
                                     limit_choices_to=models.Q(app_label='rayures'))
    content_object = GenericForeignKey('content_type', 'id')
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    deleted_at = models.DateTimeField(null=True)

    event = models.ForeignKey('rayures.Event', on_delete=models.DO_NOTHING, null=True)
    idempotency_key = models.TextField(null=True)
    request_id = models.TextField(null=True)
    source = models.TextField(null=True)

    def __str__(self):
        return f'{self.content_type} {self.id}'



# Generated by Django 2.0.6 on 2018-07-01 18:32

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Customer',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Plan',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('object_id', models.CharField(default=None, max_length=50, null=True)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
                ('content_type', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.SET_DEFAULT, to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Card',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Charge',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Coupon',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='InvoiceItem',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SKU',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Source',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Transfer',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BalanceTransaction',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Dispute',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Refund',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BankAccount',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Payout',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Application',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='IssuerFraudRecord',
            fields=[
                ('id', models.CharField(editable=False, max_length=255, primary_key=True, serialize=False)),
                ('api_version', models.CharField(editable=False, max_length=12)),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('deleted_at', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RayuresMeta',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('deleted_at', models.DateTimeField(null=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('idempotency_key', models.TextField(null=True)),
                ('request_id', models.TextField(null=True)),
                ('source', models.TextField(null=True)),
                ('event', models.ForeignKey(null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='rayures.Event')),
            ],
        ),
        migrations.CreateModel(
            name='RayureEventProcess',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(choices=[('received', 'received'), ('success', 'success'), ('failure', 'failure')], default='received', max_length=12)),
                ('started_at', models.DateTimeField(auto_now_add=True)),
                ('ended_at', models.DateTimeField(editable=False, null=True)),
                ('event', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='rayures.Event')),
                ('traces', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
            ],
            options={
                'verbose_name': 'Rayure event process',
                'verbose_name_plural': 'Rayure event processes',
            },
        ),
        migrations.CreateModel(
            name='RayureEventProcessingError',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('http_body', models.TextField()),
                ('func', models.TextField(null=True)),
                ('message', models.TextField()),
                ('traceback', models.TextField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('acknowledged_at', models.DateTimeField(null=True)),
                ('process', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='errors', to='rayures.RayureEventProcess')),
            ],
        ),
    ]



import stripe
from django.core.management.base import BaseCommand
from rayures.reconciliation import reconciliate_event


class Command(BaseCommand):
    help = 'Sync stripe events'

    def handle(self, *args, **options):
        for obj in stripe.Event.all(limit=1000).auto_paging_iter():
            obj, created = reconciliate_event(obj, persist=True)
            status = 'created' if created else 'updated'
            self.stdout.write(f'{type(obj)._stripe_object} {obj.id} {status}')



import json
import logging
import stripe
from .events import dispatch
from .exceptions import DispatchException
from .models import RayureEventProcess
from .reconciliation import reconciliate_event
from datetime import timedelta
from django.apps import apps
from django.db.models import Q
from django.db.models.functions import Now
from django.http import HttpResponse, JsonResponse
from django.views.decorators.http import require_http_methods
from django.views.decorators.csrf import csrf_exempt


logger = logging.getLogger('rayures')


@csrf_exempt
@require_http_methods(["POST"])
def stripe_web_hook(request):
    """Handle stripe webhooks
    """
    if request.method != 'POST':
        return HttpResponse(status=400)

    try:
        response = load_stripe_event(request)
        if isinstance(response, HttpResponse):
            return response
        state = response
    except ValueError as error:
        # Invalid payload
        logger.error('stripe_web_hook ValueError - error: %s' % error)
        return HttpResponse(status=400)
    except stripe.error.SignatureVerificationError as error:
        # Invalid signature
        logger.error('stripe_web_hook SignatureVerificationError - error: %s' % error)
        return HttpResponse(status=400)
    data = handle_dispatch(state)
    if isinstance(data, HttpResponse):
        return data
    return JsonResponse(data, status=200)


def load_stripe_event(request) -> stripe.StripeObject:
    endpoint_secret = apps.app_configs['rayures'].endpoint_secret
    payload = request.body
    if endpoint_secret:
        # verify signature
        sig_header = request.META.get('HTTP_STRIPE_SIGNATURE', None)
        if not sig_header:
            return HttpResponse(status=400)
        try:
            return stripe.Webhook.construct_event(
                payload, sig_header, endpoint_secret)
        except ValueError as error:
            # Invalid payload
            logger.error('stripe_web_hook ValueError - error: %s' % error)
            return HttpResponse(status=400)
        except stripe.error.SignatureVerificationError as error:
            # Invalid signature
            logger.error('stripe_web_hook SignatureVerificationError - error: %s' % error)
            return HttpResponse(status=400)
    else:
        return json.loads(payload)


def handle_dispatch(state):
    if state['type'] == 'ping':
        return HttpResponse()
    # do we have a running process for this event?
    predicate = Q(
        event_id=state['id'],
        status='received',
        started_at__gte=Now() - timedelta(minutes=15)  # staled event for whatever reason
    )
    process_id = RayureEventProcess.objects \
        .filter(predicate) \
        .values_list('id', flat=True) \
        .first()
    if process_id:
        # discard processing
        return {'id': process_id, 'status': 'running'}
    try:
        event, created = reconciliate_event(state, persist=True)
        process = dispatch(event)
    except DispatchException as error:
        data = {'errors': error.formatted_errors}
        process = error.process
    else:
        data = {}
    data.update({'id': process.id, 'status': process.status, 'traces': process.traces})
    return data



import logging
from .events import listen
from .models import PersistedModel

logger = logging.getLogger('rayures')


@listen('*', position=10)
def persist_obj(event, obj):
    try:
        if isinstance(obj, PersistedModel):
            obj.save()
            logger.info(f'persisted {obj}', extra={'obj': obj})
        else:
            logger.info(f'ephemeral {obj}', extra={'obj': obj})
    except Exception as error:
        logger.error(f'failed to persist {obj.id}: {error}', extra={'obj': obj})



from .utils import price_from_stripe, dt_from_stripe
from django.db import models
from django.db.models.expressions import Col
from django.db.models.lookups import IExact
from django.db.models.fields.related_lookups import RelatedLookupMixin
from django.db.models.signals import post_init
from django.contrib.postgres.fields import JSONField as JSONFieldBase

__all__ = ['IntegerField']


class DatetimeProxy:
    def __init__(self, source, field_name):
        self.source = source
        self.path = source.split('.')
        self.field_name = field_name

    def __get__(self, obj, type=None):
        if obj is None:
            return obj
        try:
            value = obj.data
            for p in self.path:
                value = value.get(p, MISSING)
                if value is MISSING:
                    return
        except AttributeError:
            return
        if value is not None:
            return dt_from_stripe(value)


class CharProxy:
    def __init__(self, source, field_name):
        self.source = source
        self.path = source.split('.')
        self.field_name = field_name

    def __get__(self, obj, type=None):
        if obj is None:
            return obj
        try:
            value = obj.data
            for p in self.path:
                value = value.get(p, MISSING)
                if value is MISSING:
                    return
        except AttributeError:
            return
        if value is not None:
            return str(value)


class IntegerProxy:
    def __init__(self, source, field_name):
        self.source = source
        self.path = source.split('.')
        self.field_name = field_name

    def __get__(self, obj, type=None):
        if obj is None:
            return obj
        try:
            value = obj.data
            for p in self.path:
                value = value.get(p, MISSING)
                if value is MISSING:
                    return
        except AttributeError:
            return
        if value is not None:
            return int(value)


class PriceProxy:
    def __init__(self, source, field_name):
        self.source = source
        self.path = source.split('.')
        self.currency_path = self.path[:-1] + ['currency']
        self.field_name = field_name

    def __get__(self, obj, type=None):
        if obj is None:
            return obj
        try:
            value = obj.data
            for p in self.path:
                value = value.get(p, MISSING)
                if value is MISSING:
                    return
            currency = obj.data
            for p in self.currency_path:
                currency = currency.get(p, MISSING)
                if currency is MISSING:
                    return
        except AttributeError:
            return

        if value is not None:
            # TODO: convert the value to local
            return price_from_stripe(value, currency)


class BooleanProxy:
    def __init__(self, source, field_name):
        self.source = source
        self.path = source.split('.')
        self.field_name = field_name

    def __get__(self, obj, type=None):
        if obj is None:
            return obj
        try:
            value = obj.data
            for p in self.path:
                value = value.get(p, MISSING)
                if value is MISSING:
                    return
        except AttributeError:
            return
        if value is not None:
            return bool(value)


class HashProxy:
    def __init__(self, source, field_name):
        self.source = source
        self.path = source.split('.')
        self.field_name = field_name

    def __get__(self, obj, type=None):
        if obj is None:
            return obj
        try:
            value = obj.data
            for p in self.path:
                value = value.get(p, MISSING)
                if value is MISSING:
                    return
        except AttributeError:
            return
        return value


class StripeCol(Col):
    def as_sql(self, compiler, connection):
        qn = compiler.quote_name_unless_alias
        *prev, last = ["data"] + [f"'{p}'" for p in self.target.source.split('.')]
        field = '->'.join(prev) + '->>' + last
        # cast for now
        field = "%s.%s" % (qn(self.alias), field)
        if isinstance(self.target, DateTimeField):
            field = f'to_timestamp(({field})::text::double precision)'
        elif isinstance(self.target, IntegerField):
            field = f'({field})::text::numeric'
        elif isinstance(self.target, BooleanField):
            field = f'({field})::text::bool'
        elif isinstance(self.target, HashField):
            field = '->'.join(prev) + '->' + last
            field = "%s.%s" % (qn(self.alias), field)
        else:
            field = f'({field})::text'
        return field, []


class StripeField(models.Field):
    proxy = None

    def __init__(self, *args, source, **kwargs):
        """
        Parameters:
            source (str): the path in data JSON
        """
        self.source = source
        kwargs['null'] = True
        kwargs['editable'] = False
        kwargs['serialize'] = False
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["source"] = self.source
        del kwargs["editable"]
        del kwargs["serialize"]
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name):
        self.name = name
        self.verbose_name = name.replace('_', ' ')
        self.field_name = name
        self.attname = name
        self.model = cls

        self.concrete = False
        self.column = f'__{self.source}__'
        self.proxy = type(self).proxy(self.source, self.attname)
        cls._meta.add_field(self, private=True)
        if not getattr(cls, self.attname, None):
            setattr(cls, self.attname, self.proxy)
        if not cls._meta.abstract:
            post_init.connect(self.rebound_fields, sender=cls)

    def rebound_fields(self, instance, *args, **kwargs):
        self.rebound(instance)

    def rebound(self, instance):
        value = self.proxy.__get__(instance)
        setattr(instance, self.attname, value)
        pass

    def get_col(self, alias, output_field=None):
        col = super().get_col(alias, output_field)
        if isinstance(col, Col):
            col.__class__ = StripeCol
        return col

    # def select_format(self, compiler, sql, params):
    #     sql, params = super().select_format(compiler, sql, params)
    #     print('select_format', self, sql, params)
    #     return sql, params

    # def get_lookup(self, name):
    #     result = super().get_lookup(name)
    #     print('get_lookup', self, result, name)
    #     # get_lookup rayures.Coupon.created_at <class 'django.db.models.lookups.GreaterThanOrEqual'> gte
    #     # get_lookup rayures.Coupon.created_at <class 'django.db.models.lookups.LessThan'> lt
    #     return result

    # def get_transform(self, name):
    #     result = super().get_transform(name)
    #     print('get_transform', self, result, name)
    #     return result


class IntegerField(StripeField, models.IntegerField):
    # description = _("String (up to %(max_length)s)")
    proxy = IntegerProxy

    def get_internal_type(self):
        return 'IntegerField'


class CharField(StripeField, models.CharField):
    # description = _("String (up to %(max_length)s)")
    proxy = CharProxy

    # def __init__(self, *args, **kwargs):
    #     """
    #     Parameters:
    #         source (str): the path in data JSON
    #     """
    #     kwargs['max_length'] = 2000
    #     super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return 'CharField'


class DateTimeField(StripeField, models.DateTimeField):
    # description = _("String (up to %(max_length)s)")
    proxy = DatetimeProxy

    def get_internal_type(self):
        return 'DateTimeField'


class PriceField(StripeField):
    # description = _("String (up to %(max_length)s)")
    proxy = PriceProxy


class BooleanField(StripeField, models.NullBooleanField):
    # description = _("String (up to %(max_length)s)")
    proxy = BooleanProxy

    def get_internal_type(self):
        return 'NullBooleanField'


class HashField(StripeField, JSONFieldBase):
    # description = _("String (up to %(max_length)s)")
    proxy = HashProxy

    def get_internal_type(self):
        return 'JSONField'


MISSING = object()


class ForeignKey(models.ForeignKey):
    def __init__(self, to, related_name=None, related_query_name=None,
                 limit_choices_to=None, parent_link=False, to_field=None,
                 source=None,
                 **kwargs):
        kwargs['to'] = to
        kwargs['related_name'] = related_name
        kwargs['related_query_name'] = related_query_name
        kwargs['limit_choices_to'] = limit_choices_to
        kwargs['parent_link'] = parent_link
        kwargs['to_field'] = to_field
        # forced
        kwargs['db_constraint'] = False
        kwargs['db_index'] = False
        kwargs['null'] = True
        kwargs['default'] = None
        kwargs['on_delete'] = models.DO_NOTHING

        # our
        kwargs['editable'] = False
        self.source = source

        super().__init__(**kwargs)

    def get_col(self, alias, output_field=None):
        col = super().get_col(alias, output_field)
        if isinstance(col, Col):
            col.__class__ = StripeCol
        return col

    def contribute_to_class(self, cls, name, private_only=False, **kwargs):
        return super().contribute_to_class(cls, name, private_only=True, **kwargs)


@ForeignKey.register_lookup
class RelatedIExact(RelatedLookupMixin, IExact):
    pass
